# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import properties

from .. import BaseModel


class Workflow(BaseModel):

    mailbox_id = properties.Integer(
        'The mailbox ID that this workflow is associated with.',
        required=True,
    )
    type = properties.StringChoice(
        'The type of workflow.',
        choices=['automatic', 'manual'],
        default='manual',
        required=True,
    )
    status = properties.StringChoice(
        'The status of the workflow.',
        choices=['active', 'inactive', 'invalid'],
        default='invalid',
        required=True,
    )
    order = properties.Integer(
        'The order of the workflow.',
        default=1,
        required=True,
    )
    name = properties.String(
        'Workflow name.',
        required=True,
    )
    created_at = properties.DateTime(
        'UTC time when this workflow was created.',
    )
    modified_at = properties.DateTime(
        'UTC time when this workflow was modified.',
    )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

from .. import BaseApi

from ..models.tag import Tag


class Tags(BaseApi):
    """This represents the ``Tags`` Endpoint.

    The following aspects are implemented:

    * `List Tags
      <http://developer.helpscout.net/help-desk-api/tags/list/>`_
      (:func:`helpscout.apis.tags.Tags.list`)
    """

    __object__ = Tag

    @classmethod
    def list(cls, session):
        """List the tags.

        Args:
            session (requests.sessions.Session): Authenticated session.

        Returns:
            RequestPaginator(output_type=helpscout.models.Tag): Tags iterator.
        """
        return cls('/tags.json', session=session)



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

from .address import Address
from .attachment import Attachment
from .attachment_data import AttachmentData
from .base_conversation import BaseConversation
from .chat import Chat
from .conversation import Conversation
from .custom_field import CustomField
from .customer import Customer
from .email import Email
from .field import Field
from .folder import Folder
from .mailbox import Mailbox
from .mailbox_ref import MailboxRef
from .option import Option
from .person import Person
from .phone import Phone
from .rating import Rating
from .search_conversation import SearchConversation
from .search_customer import SearchCustomer
from .social_profile import SocialProfile
from .source import Source
from .tag import Tag
from .thread import Thread
from .user import User
from .website import Website
from .workflow import Workflow


__all__ = [
    'Address',
    'Attachment',
    'AttachmentData',
    'BaseConversation',
    'Chat',
    'Conversation',
    'CustomField',
    'Customer',
    'Email',
    'Field',
    'Folder',
    'Mailbox',
    'MailboxRef',
    'Option',
    'Person',
    'Phone',
    'Rating',
    'SearchConversation',
    'SearchCustomer',
    'SocialProfile',
    'Source',
    'Tag',
    'Thread',
    'User',
    'Website',
    'Workflow',
]



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

from .. import BaseApi

from ..models.customer import Customer
from ..models.search_customer import SearchCustomer

from ..request_paginator import RequestPaginator


class Customers(BaseApi):
    """This represents the ``Customers`` Endpoint.

    The following aspects are implemented:

    * `List Customers
      <http://developer.helpscout.net/help-desk-api/customers/list/>`_
      (:func:`helpscout.apis.customers.Customers.list`)
    * `Search Customers
      <http://developer.helpscout.net/help-desk-api/search/customers/>`_
      (:func:`helpscout.apis.customers.Customers.search`)
    * `Get Customer
      <http://developer.helpscout.net/help-desk-api/customers/get/>`_
      (:func:`helpscout.apis.customers.Customers.get`)
    * `Create Customer
      <http://developer.helpscout.net/help-desk-api/customers/create/>`_
      (:func:`helpscout.apis.customers.Customers.create`)
    * `Update Customer
      <http://developer.helpscout.net/help-desk-api/customers/update/>`_
      (:func:`helpscout.apis.customers.Customers.update`)
    """

    __object__ = Customer

    @classmethod
    def create(cls, session, customer):
        """Create a customer.

        Args:
            session (requests.sessions.Session): Authenticated session.
            customer (helpscout.models.Customer): The customer to be created.

        Returns:
            helpscout.models.Customer: Newly created customer.
        """
        data = {
            'customer': customer.to_api(),
            'reload': True,
        }
        return cls(
            '/customers.json',
            data=data,
            request_type=RequestPaginator.POST,
            singleton=True,
            session=session,
        )

    @classmethod
    def get(cls, session, customer_id):
        """Return a specific customer.

        Args:
            session (requests.sessions.Session): Authenticated session.
            customer_id (int): The ID of the customer to get.

        Returns:
            helpscout.models.Customer: A customer singleton, if existing.
                Otherwise ``None``.
        """
        return cls(
            '/customers/%d.json' % customer_id,
            singleton=True,
            session=session,
        )

    @classmethod
    def list(cls, session, first_name=None, last_name=None, email=None,
             modified_since=None):
        """List the customers.

        Customers can be filtered on any combination of first name, last name,
        email, and modifiedSince.

        Args:
            session (requests.sessions.Session): Authenticated session.
            first_name (str, optional): First name of customer.
            last_name (str, optional): Last name of customer.
            email (str, optional): Email address of customer.
            modified_since (datetime.datetime, optional): If modified after
                this date.

        Returns:
            RequestPaginator(output_type=helpscout.models.Customer): Customers
                iterator.
        """
        data = cls.__object__.get_non_empty_vals({
            'firstName': first_name,
            'lastName': last_name,
            'email': email,
            'modifiedSince': modified_since,
        })
        return cls('/customers.json', data, session=session)

    @classmethod
    def search(cls, session, queries):
        """Search for a customer given a domain.

        Args:
            session (requests.sessions.Session): Authenticated session.
            queries (helpscout.models.Domain or iter): The queries for the
                domain. If a ``Domain`` object is provided, it will simply be
                returned. Otherwise, a ``Domain`` object will be generated
                from the complex queries. In this case, the queries should
                conform to the interface in
                :func:`helpscout.domain.Domain.from_tuple`.

        Returns:
            RequestPaginator(output_type=helpscout.models.SearchCustomer):
                SearchCustomer iterator.
        """
        domain = cls.get_search_domain(queries)
        return cls(
            '/search/customers.json',
            data={'query': str(domain)},
            session=session,
            out_type=SearchCustomer,
        )

    @classmethod
    def update(cls, session, customer):
        """Update a customer.

        Args:
            session (requests.sessions.Session): Authenticated session.
            customer (helpscout.models.Customer): The customer to be updated.

        Returns:
            helpscout.models.Customer: Freshly updated customer.
        """
        data = customer.to_api()
        data['reload'] = True
        return cls(
            '/customers/%d.json' % customer.id,
            data=data,
            request_type=RequestPaginator.PUT,
            singleton=True,
            session=session,
        )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import properties

from .base_conversation import BaseConversation


class SearchConversation(BaseConversation):
    """This represents a conversation as returned by search results."""

    mailbox_id = properties.Integer(
        'The ID of the mailbox this conversation is in.',
        required=True,
    )
    customer_name = properties.String(
        'Name of the customer this conversation is regarding.',
        required=True,
    )
    customer_email = properties.String(
        'Email address of the customer',
        required=True,
    )
    has_attachments = properties.Bool(
        '``True`` when the conversation has at least one attachment.',
    )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

from .base_model import BaseModel
from .domain import Domain
from .request_paginator import RequestPaginator


class BaseApi(object):
    """This is the API interface object to be implemented by API adapters.

    It acts as a collection for the API object that it represents, passing
    through iteration to the API's request paginator.

    Attributes:
        BASE_URI (str): HelpScout API URI base.
        paginator (RequestPaginator): Object to use for producing an iterator
         representing multiple requests (API response pages). Created on init.
        __object__ (helpscout.models.BaseModel): Model object that API
         represents.
    """

    BASE_URI = 'https://api.helpscout.net/v1'

    # This should be replaced in child classes with the correct model.
    __object__ = BaseModel

    # This is set within new, after the object has been created.
    paginator = None

    def __new__(cls, endpoint, data=None,
                request_type=RequestPaginator.GET, singleton=False,
                session=None, out_type=None):
        """Create a new API object.

        Args:
            endpoint (str): The API endpoint that this represents.
                ``BASE_URI`` will be prepended, with no slashes added.
            data (dict, optional): Data to send with the request.
            request_type (str, optional): Type of request (``GET or ``POST``).
                Defaults to ``GET``.
            singleton (bool, optional): Set this to ``True`` to assert that
                there is not more than one result, and return the first result
                (or ``None`` if there is no result).
            session (requests.Session, optional): An authenticated requests
                session to use.
            out_type (BaseModel, optional): If set, this object will be used
                for the creation of the models, instead of the one set in
                ``cls.__object__``.

        Raises:
            HelpScoutRemoteException: If ``singleton`` is ``True``, but the
                remote API responds with more than one result.

        Returns:
            BaseApi: An instance of an API object, if ``singleton`` is
                ``False``.
            BaseModel: An instance of a Model, if ``singleton`` is
                ``True`` and there are results.
            None: If ``singleton`` is ``True`` and there are no results.
        """
        if out_type is None:
            out_type = cls.__object__
        paginator = RequestPaginator(
            endpoint='%s%s' % (cls.BASE_URI, endpoint),
            data=data,
            output_type=out_type.from_api,
            request_type=request_type,
            session=session,
        )
        if singleton:
            results = paginator.call(paginator.data)
            result_length = len(results)
            if result_length == 0:
                return None
            return out_type.from_api(**results[0])
        obj = super(BaseApi, cls).__new__(cls)
        obj.paginator = paginator
        return obj

    def __iter__(self):
        """Pass through iteration to the API response."""
        for row in self.paginator:
            yield row
        raise StopIteration()

    @classmethod
    def new_object(cls, data):
        """Return a new object of the correct type from the data.

        Args:
            data (dict): Data dictionary that should be converted to an
                object. It can use either camelCase keys or snake_case.

        Returns:
            BaseModel: A model of the type declared in ``cls.__object__``.
        """
        return cls.__object__.from_api(**data)

    @staticmethod
    def get_search_domain(queries):
        """Helper method to create search domains if needed.

        Args:
            queries (Domain or iter): The queries for the domain. If a
                ``Domain`` object is provided, it will simply be returned.
                Otherwise, a ``Domain`` object will be generated from the
                complex queries. In this case, the queries should conform
                to the interface in
                :func:`helpscout.domain.Domain.from_tuple`.

        Returns:
            Domain: Either the provided domain, or one created from
                ``queries``.
        """
        if isinstance(queries, Domain):
            return queries
        return Domain.from_tuple(queries)



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import properties

from .. import BaseModel
from ..models import Conversation, Customer, Rating


class WebHookEvent(properties.HasProperties):
    """``WebHookEvent`` represents an authenticated web hook request.

    Note that this object is meant to represent an authenticated Web Hook,
    and therefore is completely naive of all things authentication. Any
    request authentication/validation should happen in the ``WebHook``.
    """

    # Map the event prefixes to their corresponding data models.
    EVENT_PREFIX_TO_MODEL = {
        'convo': Conversation,
        'customer': Customer,
        'satisfaction': Rating,
    }

    event_type = properties.String(
        'The event type that this object represents.',
        required=True,
    )
    record = properties.Instance(
        'The parsed data record that was received in the request.',
        instance_class=BaseModel,
        required=True,
    )

    def __init__(self, *args, **kwargs):
        """Parse raw record data if required.

        Args:
            record (dict or BaseModel): The record data that was received for
                the request. If it is a ``dict``, the data will be parsed
                using the proper model's ``from_api`` method.
        """
        if isinstance(kwargs.get('record'), dict):
            prefix, _ = kwargs['event_type'].split('.', 1)
            model = self.EVENT_PREFIX_TO_MODEL[prefix]
            kwargs['record'] = model.from_api(**kwargs['record'])
        super(WebHookEvent, self).__init__(*args, **kwargs)



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import properties

from .. import BaseModel


class Person(BaseModel):
    """This is a subset of the data representing a Customer, User or Team.

    The ``type`` property will specify if this person is represented by a
    ``user``, ``customer`` or ``team``.
    """

    full_name = properties.String(
        'Full name for the customer',
    )
    first_name = properties.String(
        'First name',
        required=True,
    )
    last_name = properties.String(
        'Last name',
        required=True,
    )
    email = properties.String(
        'Email',
    )
    phone = properties.String(
        'Phone is only populated when the Person is a customer associated '
        'with a Conversation of type ``phone`` and a phone number was '
        'specified at the time the conversation was created.',
    )
    type = properties.StringChoice(
        'The type of person.',
        choices=['user', 'customer', 'team'],
        default='customer',
        required=True,
    )
    photo_url = properties.String(
        'The user\'s photo, if one exists.',
    )
    created_at = properties.DateTime(
        'UTC time when this customer was created.',
    )
    modified_at = properties.DateTime(
        'UTC time when this customer was modified.',
    )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

from .. import BaseApi

from ..models.conversation import Conversation
from ..models.search_conversation import SearchConversation
from ..models.thread import Thread

from ..request_paginator import RequestPaginator


class Conversations(BaseApi):
    """This represents the ``Conversations`` Endpoint.

    The following aspects are implemented:

    * `List Conversations
      <http://developer.helpscout.net/help-desk-api/conversations/list/>`_
      (:func:`helpscout.apis.conversations.Conversations.list`)
    * `Search Conversations
      <http://developer.helpscout.net/help-desk-api/search/conversations/>`_
      (:func:`helpscout.apis.conversations.Conversations.search`)
    * `Get Conversation
      <http://developer.helpscout.net/help-desk-api/conversations/get/>`_
      (:func:`helpscout.apis.conversations.Conversations.get`)
    * `Create Conversation
      <http://developer.helpscout.net/help-desk-api/conversations/create/>`_
      (:func:`helpscout.apis.conversations.Conversations.create`)
    * `Update Conversation
      <http://developer.helpscout.net/help-desk-api/conversations/update/>`_
      (:func:`helpscout.apis.conversations.Conversations.update`)
    * `Delete Conversation
      <http://developer.helpscout.net/help-desk-api/conversations/delete/>`_
      (:func:`helpscout.apis.conversations.Conversations.delete`)
    * `Create Thread
      <http://developer.helpscout.net/help-desk-api/conversations/
      create-thread/>`_
      (:func:`helpscout.apis.conversations.Conversations.create_thread`)
    * `Update Thread
      <http://developer.helpscout.net/help-desk-api/conversations/
      update-thread/>`_
      (:func:`helpscout.apis.conversations.Conversations.update_thread`)
    """

    __object__ = Conversation

    @classmethod
    def create(cls, session, conversation, imported=False, auto_reply=False):
        """Create a conversation.

        Please note that conversation cannot be created with more than 100
        threads, if attempted the API will respond with HTTP 412.

        Args:
            session (requests.sessions.Session): Authenticated session.
            conversation (helpscout.models.Conversation): The conversation
             to be created.
            imported (bool, optional): The ``imported`` request parameter
             enables conversations to be created for historical purposes (i.e.
             if moving from a different platform, you can import your
             history). When ``imported`` is set to ``True``, no outgoing
             emails or notifications will be generated.
            auto_reply (bool): The ``auto_reply`` request parameter enables
             auto replies to be sent when a conversation is created via the
             API. When ``auto_reply`` is set to ``True``, an auto reply will
             be sent as long as there is at least one ``customer`` thread in
             the conversation.

        Returns:
            helpscout.models.Conversation: Newly created conversation.
        """
        data = {
            'conversation': conversation.to_api(),
            'imported': imported,
            'auto_reply': auto_reply,
            'reload': True,
        }
        return cls(
            '/conversations.json',
            data=data,
            request_type=RequestPaginator.POST,
            singleton=True,
            session=session,
        )

    @classmethod
    def create_thread(cls, session, conversation, thread, imported=False):
        """Create a conversation thread.

        Please note that threads cannot be added to conversations with 100
        threads (or more), if attempted the API will respond with HTTP 412.

        Args:
            conversation (helpscout.models.Conversation): The conversation
             that the thread is being added to.
            session (requests.sessions.Session): Authenticated session.
            thread (helpscout.models.Thread): The thread to be created.
            imported (bool, optional): The ``imported`` request parameter
             enables conversations to be created for historical purposes (i.e.
             if moving from a different platform, you can import your
             history). When ``imported`` is set to ``True``, no outgoing
             emails or notifications will be generated.

        Returns:
            helpscout.models.Thread: Newly created thread.
        """
        data = {
            'thread': thread.to_api(),
            'imported': imported,
            'reload': True,
        }
        return cls(
            '/conversations/%s.json' % conversation.id,
            data=data,
            request_type=RequestPaginator.POST,
            singleton=True,
            session=session,
            out_type=Thread,
        )

    @classmethod
    def delete(cls, session, conversation):
        """Delete a conversation.

        Args:
            session (requests.sessions.Session): Authenticated session.
            conversation (helpscout.models.Conversation): The conversation to
                be deleted.

        Returns:
            bool: Status
        """
        return cls(
            '/conversations/%s.json' % conversation.id,
            request_type=RequestPaginator.DELETE,
            singleton=True,
            session=session,
        )

    @classmethod
    def find_customer(cls, session, mailbox, customer):
        """Return conversations for a specific customer in a mailbox.

        Args:
            session (requests.sessions.Session): Authenticated session.
            mailbox (helpscout.models.Mailbox): Mailbox to search.
            customer (helpscout.models.Customer): Customer to search for.

        Returns:
            RequestPaginator(output_type=helpscout.models.Conversation):
                Conversations iterator.
        """
        return cls(
            '/mailboxes/%d/customers/%s/conversations.json' % (
                mailbox.id, customer.id,
            ),
            session=session,
        )

    @classmethod
    def find_user(cls, session, mailbox, user):
        """Return conversations for a specific user in a mailbox.

        Args:
            session (requests.sessions.Session): Authenticated session.
            mailbox (helpscout.models.Mailbox): Mailbox to search.
            user (helpscout.models.User): User to search for.

        Returns:
            RequestPaginator(output_type=helpscout.models.Conversation):
                Conversations iterator.
        """
        return cls(
            '/mailboxes/%d/users/%s/conversations.json' % (
                mailbox.id, user.id,
            ),
            session=session,
        )

    @classmethod
    def get(cls, session, conversation_id):
        """Return a specific conversation.

        Args:
            session (requests.sessions.Session): Authenticated session.
            conversation_id (int): The ID of the conversation to get.

        Returns:
            helpscout.models.Conversation: A conversation singleton, if
                existing. Otherwise ``None``.
        """
        return cls(
            '/conversations/%d.json' % conversation_id,
            singleton=True,
            session=session,
        )

    @classmethod
    def list(cls, session, mailbox):
        """Return conversations in a mailbox.

        Args:
            session (requests.sessions.Session): Authenticated session.
            mailbox (helpscout.models.Mailbox): Mailbox to list.

        Returns:
            RequestPaginator(output_type=helpscout.models.Conversation):
                Conversations iterator.
        """
        return cls(
            '/mailboxes/%d/conversations.json' % mailbox.id,
            session=session,
        )

    @classmethod
    def list_folder(cls, session, mailbox, folder):
        """Return conversations in a specific folder of a mailbox.

        Args:
            session (requests.sessions.Session): Authenticated session.
            mailbox (helpscout.models.Mailbox): Mailbox that folder is in.
            folder (helpscout.models.Folder): Folder to list.

        Returns:
            RequestPaginator(output_type=helpscout.models.Conversation):
                Conversations iterator.
        """
        return cls(
            '/mailboxes/%d/folders/%s/conversations.json' % (
                mailbox.id, folder.id,
            ),
            session=session,
        )

    @classmethod
    def search(cls, session, queries):
        """Search for a conversation given a domain.

        Args:
            session (requests.sessions.Session): Authenticated session.
            queries (helpscout.models.Domain or iter): The queries for the
                domain. If a ``Domain`` object is provided, it will simply be
                returned. Otherwise, a ``Domain`` object will be generated
                from the complex queries. In this case, the queries should
                conform to the interface in
                :func:`helpscout.domain.Domain.from_tuple`.

        Returns:
            RequestPaginator(output_type=helpscout.models.SearchCustomer):
                SearchCustomer iterator.
        """
        domain = cls.get_search_domain(queries)
        return cls(
            '/search/conversations.json',
            data={'query': str(domain)},
            session=session,
            out_type=SearchConversation,
        )

    @classmethod
    def update(cls, session, conversation):
        """Update a conversation.

        Args:
            session (requests.sessions.Session): Authenticated session.
            conversation (helpscout.models.Conversation): The conversation to
                be updated.

        Returns:
            helpscout.models.Conversation: Freshly updated conversation.
        """
        data = conversation.to_api()
        data['reload'] = True
        return cls(
            '/conversations/%s.json' % conversation.id,
            data=data,
            request_type=RequestPaginator.PUT,
            singleton=True,
            session=session,
        )

    @classmethod
    def update_thread(cls, session, conversation, thread):
        """Update a thread.

        Args:
            session (requests.sessions.Session): Authenticated session.
            conversation (helpscout.models.Conversation): The conversation
                that the thread belongs to.
            thread (helpscout.models.Thread): The thread to be updated.

        Returns:
            helpscout.models.Thread: Freshly updated thread.
        """
        data = thread.to_api()
        data['reload'] = True
        return cls(
            '/conversations/%s/threads/%d.json' % (
                conversation.id, thread.id,
            ),
            data=data,
            request_type=RequestPaginator.PUT,
            singleton=True,
            session=session,
            out_type=Thread,
        )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import properties

from .person import Person


class User(Person):

    role = properties.StringChoice(
        'Role of this user.',
        choices=['owner', 'admin', 'user'],
        required=True,
    )
    timezone = properties.String(
        'Name of the user\'s time zone.',
    )
    type = properties.StringChoice(
        'The type of user.',
        choices=['user', 'team'],
        required=True,
    )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import properties
import unittest

from datetime import datetime

from .. import BaseModel
from ..exceptions import HelpScoutValidationException


class TestModel(BaseModel):
    a_key = properties.String('A key')
    sub_instance = properties.Instance(
        'Sub Instance', instance_class=BaseModel,
    )
    list = properties.List(
        'List', prop=properties.Instance('List Instance',
                                         instance_class=BaseModel),
    )
    date = properties.DateTime('DateTime')
    not_a_field = True


class TestBaseModel(unittest.TestCase):

    def setUp(self):
        super(TestBaseModel, self).setUp()
        self.test_values = {
            'aKey': 'value',
            'subInstance': {
                'id': 1234,
            },
            'list': [
                {'id': 4321},
            ],
            'date': datetime(2017, 1, 1, 0, 0, 0),
        }
        self.internal_values = {
            'a_key': self.test_values['aKey'],
            'sub_instance': self.test_values['subInstance'],
            'list': self.test_values['list'],
            'date': self.test_values['date'],
        }

    def new_record(self):
        return TestModel.from_api(**self.test_values)

    def test_from_api_new_instance(self):
        """It should return a new instance of the class."""
        self.assertIsInstance(self.new_record(), TestModel)

    def test_from_api_camel_to_snake(self):
        """It should call the init with snake cased keys."""
        self.assertEqual(self.new_record().a_key, self.test_values['aKey'])

    def test_from_api_invalid_attribute(self):
        """It should not allow invalid attribute assignments."""
        with self.assertRaises(HelpScoutValidationException):
            BaseModel.from_api(**{'no exist': 'value'})

    def test_from_api_sub_instance(self):
        """It should properly instantiate sub-instances."""
        self.assertIsInstance(self.new_record().sub_instance, BaseModel)

    def test_from_api_list(self):
        """It should properly convert lists of instances."""
        res = self.new_record()
        self.assertIsInstance(res.list, list)
        self.assertIsInstance(res.list[0], BaseModel)

    def test_to_api_camel_case(self):
        """It should properly camel case the API args."""
        res = self.new_record().to_api()
        self.assertEqual(res['aKey'], self.internal_values['a_key'])

    def test_to_api_instance(self):
        """It should properly convert sub-instances to dict."""
        res = self.new_record().to_api()
        self.assertIsInstance(res['subInstance'], dict)

    def test_to_api_list(self):
        """It should properly convert sub-instances within lists props."""
        res = self.new_record().to_api()
        self.assertIsInstance(res['list'], list)
        self.assertIsInstance(res['list'][0], dict)

    def test_to_api_date(self):
        """It should return the serialized datetime."""
        self.assertEqual(
            self.new_record().to_api()['date'], '2017-01-01T00:00:00Z',
        )

    def test_get_non_empty_vals(self):
        """It should return the dict without NoneTypes."""
        expect = {
            'good_int': 1234,
            'good_false': False,
            'good_true': True,
            'bad': None,
        }
        res = BaseModel.get_non_empty_vals(expect)
        del expect['bad']
        self.assertDictEqual(res, expect)

    def test_dict_lookup_exist(self):
        """It should return the attribute value when it exists."""
        self.assertEqual(
            self.new_record()['a_key'], self.internal_values['a_key'],
        )

    def test_dict_lookup_no_exist(self):
        """It should raise a KeyError when the attribute isn't a field."""
        with self.assertRaises(KeyError):
            self.new_record()['not_a_field']

    def test_dict_set_exist(self):
        """It should set the attribute via the items."""
        expect = 'Test-Expect'
        record = self.new_record()
        record['a_key'] = expect
        self.assertEqual(record.a_key, expect)

    def test_dict_set_no_exist(self):
        """It should raise a KeyError and not change the non-field."""
        record = self.new_record()
        with self.assertRaises(KeyError):
            record['not_a_field'] = False
        self.assertTrue(record.not_a_field)

    def test_get_exist(self):
        """It should return the attribute if it exists."""
        self.assertEqual(
            self.new_record().get('a_key'), self.internal_values['a_key'],
        )

    def test_get_no_exist(self):
        """It should return the default if the attribute doesn't exist."""
        expect = 'Test'
        self.assertEqual(
            self.new_record().get('not_a_field', expect), expect,
        )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

from .. import BaseApi

from ..models.user import User


class Users(BaseApi):
    """This represents the ``Users`` Endpoint.

    The following aspects are implemented:

    * `List Users
      <http://developer.helpscout.net/help-desk-api/users/list/>`_
      (:func:`helpscout.apis.users.Users.list`)
    * `Get Users
      <http://developer.helpscout.net/help-desk-api/users/get/>`_
      (:func:`helpscout.apis.users.Users.get`)
    * `Get Current User
      <http://developer.helpscout.net/help-desk-api/users/get/>`_
      (:func:`helpscout.apis.users.Users.get_me`)
    * `List Users By Mailbox
      <http://developer.helpscout.net/help-desk-api/users/mailbox-users/>`_
      (:func:`helpscout.apis.users.Users.find_in_mailbox`)
    """

    __object__ = User

    @classmethod
    def get(cls, session, user_id):
        """Return a specific user.

        Args:
            session (requests.sessions.Session): Authenticated session.
            user_id (int): The ID of the user to get.

        Returns:
            helpscout.models.User: A user singleton, if existing.
                Otherwise ``None``.
        """
        return cls(
            '/users/%d.json' % user_id,
            singleton=True,
            session=session,
        )

    @classmethod
    def get_me(cls, session):
        """Return the current user.

        Args:
            session (requests.sessions.Session): Authenticated session.

        Returns:
            helpscout.models.User: A user singleton
        """
        return cls(
            '/users/me.json',
            singleton=True,
            session=session,
        )

    @classmethod
    def list(cls, session):
        """List the users.

        Args:
            session (requests.sessions.Session): Authenticated session.

        Returns:
            RequestPaginator(output_type=helpscout.models.User): Users
                iterator.
        """
        return cls('/users.json', session=session)

    @classmethod
    def find_in_mailbox(cls, session, mailbox_or_id):
        """Get the users that are associated to a Mailbox.

        Args:
            session (requests.sessions.Session): Authenticated session.
            mailbox_or_id (MailboxRef or int): Mailbox of the ID of the
                mailbox to get the folders for.

        Returns:
            RequestPaginator(output_type=helpscout.models.User): Users
                iterator.
        """
        if isinstance(mailbox_or_id, User):
            mailbox_or_id = mailbox_or_id.id
        return cls(
            '/mailboxes/%d/users.json' % mailbox_or_id,
            session=session,
        )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import mock
import unittest

from datetime import datetime

from ..domain import Domain


class TestDomain(unittest.TestCase):

    def test_init_adds_queries(self):
        """It should add queries if defined."""
        with mock.patch.object(Domain, 'add_query') as add:
            Domain(['expect'], 'test')
            add.assert_called_once_with('expect', 'test')

    def test_from_tuple(self):
        """It should return the proper domain."""
        data = [('status', 'active'),
                ('number', 1234),
                ('modified_at', datetime(2017, 1, 1), datetime(2017, 1, 2)),
                ('subject', 'Test1'),
                'OR',
                ('subject', 'Test2')]
        res = Domain.from_tuple(data)
        self.assertEqual(
            str(res),
            '('
            'status:"active" '
            'AND number:1234 '
            'AND modifiedAt:[2017-01-01T00:00:00Z TO 2017-01-02T00:00:00Z] '
            'AND subject:"Test1" '
            'OR subject:"Test2"'
            ')',
        )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import properties

from .person import Person


class SearchCustomer(Person):
    """This represents a customer as returned by a search."""

    photo_type = properties.StringChoice(
        'Type of photo.',
        choices=['unknown',
                 'gravatar',
                 'twitter',
                 'facebook',
                 'googleprofile',
                 'googleplus',
                 'linkedin',
                 ],
    )
    gender = properties.StringChoice(
        'Gender of this customer.',
        choices=['female', 'male', 'unknown'],
        default='unknown',
        required=True,
    )
    age = properties.String(
        'Age (or age range) of this customer.',
    )
    organization = properties.String(
        'Company/Organization the customer identifies with.',
    )
    job_title = properties.String(
        'Job title at company/organization.',
    )
    location = properties.String(
        'Location',
    )
    emails = properties.List(
        'Email addresses for the customer.',
        prop=properties.String(''),
    )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import properties

from .. import BaseModel


class Tag(BaseModel):

    name = properties.String(
        'The tag value.',
        required=True,
    )
    slug = properties.String(
        'Slugified version of the tag value.',
    )
    count = properties.Integer(
        'The number of times the tag is used.',
    )
    color = properties.Color(
        'The tag color.',
        required=True,
    )
    created_at = properties.DateTime(
        'UTC time when this tag was created.',
    )
    modified_at = properties.DateTime(
        'UTC time when this tag was modified.',
    )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import properties

from datetime import datetime, date

from ..base_model import BaseModel


class Domain(properties.HasProperties):
    """This represents a full search query."""

    OR = 'OR'
    AND = 'AND'

    def __init__(self, queries=None, join_with=AND):
        """Initialize a domain, with optional queries."""
        self.query = []
        if queries is not None:
            for query in queries:
                self.add_query(query, join_with)

    @classmethod
    def from_tuple(cls, queries):
        """Create a ``Domain`` given a set of complex query tuples.

        Args:
            queries (iter): An iterator of complex queries. Each iteration
                should contain either:

                * A data-set compatible with :func:`~domain.Domain.add_query`
                * A string to switch the join type

                Example::

                    [('subject', 'Test1'),
                     'OR',
                     ('subject', 'Test2')',
                     ('subject', 'Test3')',
                     ]
                    # The above is equivalent to:
                    #    subject:'Test1' OR subject:'Test2' OR subject:'Test3'

                    [('modified_at', datetime(2017, 01, 01)),
                     ('status', 'active'),
                     ]
                    # The above is equivalent to:
                    #    modified_at:[2017-01-01T00:00:00Z TO *]
                    #    AND status:"active"

        Returns:
            Domain: A domain representing the input queries.
        """
        domain = cls()
        join_with = cls.AND
        for query in queries:
            if query in [cls.OR, cls.AND]:
                join_with = query
            else:
                domain.add_query(query, join_with)
        return domain

    def add_query(self, query, join_with=AND):
        """Join a new query to existing queries on the stack.

        Args:
            query (tuple or list or DomainCondition): The condition for the
                query. If a ``DomainCondition`` object is not provided, the
                input should conform to the interface defined in
                :func:`~.domain.DomainCondition.from_tuple`.
            join_with (str): The join string to apply, if other queries are
                already on the stack.
        """
        if not isinstance(query, DomainCondition):
            query = DomainCondition.from_tuple(query)
        if len(self.query):
            self.query.append(join_with)
        self.query.append(query)

    def __str__(self):
        """Return a string usable as the query in an API request."""
        return '(%s)' % ' '.join([str(q) for q in self.query])


class DomainCondition(properties.HasProperties):
    """This represents one condition of a domain query."""

    field = properties.String(
        'Field to search on',
        required=True,
    )
    value = properties.String(
        'String Value',
        required=True,
    )

    @property
    def field_name(self):
        """Return the name of the API field."""
        return BaseModel._to_camel_case(self.field)

    def __init__(self, field, value, **kwargs):
        """Initialize a new generic query condition.

        Args:
            field (str): Field name to search on. This should be the
                Pythonified name as in the internal models, not the
                name as provided in the API e.g. ``first_name`` for
                the Customer's first name instead of ``firstName``.
            value (mixed): The value of the field.
        """
        return super(DomainCondition, self).__init__(
            field=field, value=value, **kwargs
        )

    @classmethod
    def from_tuple(cls, query):
        """Create a condition from a query tuple.

        Args:
            query (tuple or list): Tuple or list that contains a query domain
                in the format of ``(field_name, field_value,
                field_value_to)``. ``field_value_to`` is only applicable in
                the case of a date search.

        Returns:
            DomainCondition: An instance of a domain condition. The specific
                type will depend on the data type of the first value provided
                in ``query``.
        """

        field, query = query[0], query[1:]

        try:
            cls = TYPES[type(query[0])]
        except KeyError:
            # We just fallback to the base class if unknown type.
            pass

        return cls(field, *query)

    def __str__(self):
        """Return a string usable as a query part in an API request."""
        return '%s:"%s"' % (self.field_name, self.value)


class DomainConditionBoolean(DomainCondition):
    """This represents an integer query."""

    value = properties.Bool(
        'Boolean Value',
        required=True,
    )

    def __str__(self):
        """Return a string usable as a query part in an API request."""
        value = 'true' if self.value else 'false'
        return '%s:%s' % (self.field_name, value)


class DomainConditionInteger(DomainCondition):
    """This represents an integer query."""

    value = properties.Integer(
        'Integer Value',
        required=True,
    )

    def __str__(self):
        """Return a string usable as a query part in an API request."""
        return '%s:%d' % (self.field_name, self.value)


class DomainConditionDateTime(DomainCondition):
    """This represents a date time query."""

    value = properties.DateTime(
        'Date From',
        required=True,
    )
    value_to = properties.DateTime(
        'Date To',
    )

    def __init__(self, field, value_from, value_to=None):
        """Initialize a new datetime query condition.

        Args:
            field (str): Field name to search on. This should be the
                Pythonified name as in the internal models, not the
                name as provided in the API e.g. ``first_name`` for
                the Customer's first name instead of ``firstName``.
            value_from (date or datetime): The start value of the field.
            value_to (date or datetime, optional): The ending value for
                the field. If omitted, will search to now.
        """
        return super(DomainConditionDateTime, self).__init__(
            field=field, value=value_from, value_to=value_to,
        )

    def __str__(self):
        """Return a string usable as a query part in an API request."""
        value_to = self.value_to.isoformat() if self.value_to else '*'
        return '%s:[%sZ TO %sZ]' % (
            self.field_name,
            self.value.isoformat(),
            value_to,
        )


TYPES = {
    bool: DomainConditionBoolean,
    int: DomainConditionInteger,
    date: DomainConditionDateTime,
    datetime: DomainConditionDateTime,
}


__all__ = [
    'Domain',
    'DomainCondition',
    'DomainConditionBoolean',
    'DomainConditionDateTime',
    'DomainConditionInteger',
]



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import properties

from .address import Address
from .chat import Chat
from .email import Email
from .phone import Phone
from .search_customer import SearchCustomer
from .social_profile import SocialProfile
from .website import Website


class Customer(SearchCustomer):
    """This represents a customer, which is a type of person."""

    background = properties.String(
        'This is the Notes field from the user interface.',
    )
    address = properties.Instance(
        'The address for this customer.',
        instance_class=Address,
    )
    social_profiles = properties.List(
        'Social profiles that represent this customer.',
        prop=SocialProfile,
    )
    emails = properties.List(
        'Email addresses for this customer.',
        prop=Email,
    )
    phones = properties.List(
        'Phone numbers for this customer.',
        prop=Phone,
    )
    chats = properties.List(
        'Chat addresses for this customer.',
        prop=Chat,
    )
    websites = properties.List(
        'Websites for this customer.',
        prop=Website,
    )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).


class HelpScoutException(Exception):
    """Base exception for HelpScout library errors."""

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return str(self.message)


class HelpScoutRemoteException(HelpScoutException):
    """Indicates that an error occurred when communicating with the remote."""

    def __init__(self, status_code, message):
        self.status_code = status_code
        super(HelpScoutRemoteException, self).__init__(message)

    def __str__(self):
        return '(%d) %s' % (self.status_code, self.message)


class HelpScoutValidationException(HelpScoutException):
    """Indicates an error while validating user-supplied data."""


class HelpScoutSecurityException(HelpScoutException):
    """Indicates a security error; probably by an invalid web hook signature.
    """



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

from .web_hook import HelpScoutWebHook
from .web_hook_event import WebHookEvent

__all__ = [
    'HelpScoutWebHook',
    'WebHookEvent',
]



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import properties

from .. import BaseModel


class Source(BaseModel):

    type = properties.StringChoice(
        'The method from which this conversation (or thread) was created.',
        choices=['email', 'web', 'notification', 'emailfwd', 'api', 'chat'],
        default='api',
        required=True,
    )
    via = properties.StringChoice(
        'Indicates what type of entity this source represents (customer, '
        'user).',
        choices=['customer', 'user'],
    )
